"""Test/sample data needed by unittests."""

import os


# Test files
TYPES_JSON_FILE = os.path.join('sample-data', 'types.json')
TYPES_JSON_GZ_FILE = os.path.join('sample-data', 'types.json.gz')
TYPES_MSG_FILE = os.path.join('sample-data', 'types.msg')
TYPES_MSG_GZ_FILE = os.path.join('sample-data', 'types.msg.gz')
TYPES_NMEA_FILE = os.path.join('sample-data', 'types.nmea')
TYPES_NMEA_GZ_FILE = os.path.join('sample-data', 'types.nmea.gz')
